from nq.filters import Filter
from nq.flags import Flags


class recurse_until:
    pass


def _common_iterable(obj):
    if isinstance(obj, dict):
        return obj
    else:
        return (index for index, value in enumerate(obj))


def _is_iterable(obj):
    return isinstance(obj, (list, tuple, set, dict))


def logger(enabled, depth):
    def log_wrapper(message, *args, **kwargs):
        if enabled:
            print(" " * depth * 2, "[nq]", message, *args, **kwargs)

    return log_wrapper


def nested_query(iterable, *path_parts, **kwargs):
    results = []
    if not _is_iterable(iterable):
        return results
    depth = kwargs["depth"] if "depth" in kwargs else 0
    verbose = kwargs["verbose"] if "verbose" in kwargs else False
    log = logger(verbose, depth)
    flags = []
    if "flags" in kwargs:
        flags = kwargs["flags"]
        if not isinstance(flags, list):
            flags = [flags]

    if len(path_parts) > 0:
        path_part = path_parts[0]
        log("Searching with path part", path_part, "on iterable", iterable)
        if isinstance(path_part, Filter):
            for key in _common_iterable(iterable):
                log("Applying filter", path_part, "on item", iterable[key])
                if path_part.check_filter(key, iterable[key], depth):
                    if len(path_parts) == 1:
                        results.append(iterable[key])
                    else:
                        results += nested_query(
                            iterable[key],
                            *path_parts[1::],
                            depth=depth + 1,
                            verbose=verbose
                        )
        elif path_part is recurse_until:
            if len(path_parts) > 1:
                test_next_part = nested_query(
                    iterable, *path_parts[1::], depth=depth + 1, verbose=verbose
                )
                if type(iterable) in (list, dict):
                    for key in _common_iterable(iterable):
                        if len(path_parts) == 1:
                            results.append(iterable[key])
                        else:
                            results += nested_query(
                                iterable[key],
                                *path_parts,
                                depth=depth + 1,
                                verbose=verbose
                            )
                if test_next_part:
                    results += test_next_part
        else:
            found_item = False
            try:
                iterable[path_part]
                found_item = True
            except Exception:
                pass
            if found_item:
                if len(path_parts) == 1:
                    results.append(iterable[path_part])
                elif type(iterable[path_part]) in (tuple, dict, list):
                    results += nested_query(
                        iterable[path_part],
                        *path_parts[1::],
                        depth=depth + 1,
                        verbose=verbose
                    )
        log("Found results", results)
        if depth == 0 and len(results) == 1:
            return results[0]
        else:
            if Flags.FIRST_RESULT in flags:
                return results[0]
            elif Flags.LAST_RESULT in flags:
                return results[-1]
            else:
                return results
    else:
        return iterable


class Tree:
    def __init__(self, data):
        self.data = data

    def query(self, *path_parts, **kwargs):
        return nested_query(self.data, *path_parts, **kwargs)



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "nq", "VERSION"), encoding="utf-8") as f:
    version = f.read()

setup(
    name="nq",
    version=version,
    description="Nested Queries",
    url="https://sgithub.fr.world.socgen/ktollec111518/nq",
    author="SG",
    author_email="",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="dictionary nested query",
    packages=find_packages(include=["nq"]),
    package_data={'': ['nq/VERSION']},
    include_package_data=True,
    install_requires=[],
    extras_require={"tests": ["pytest", "flake8", "pytest-cov"]},
    entry_points={"console_scripts": []},
    project_urls={
        "Bug Reports": "https://sgithub.fr.world.socgen/ktollec111518/nq/issues",
        "Source": "https://sgithub.fr.world.socgen/ktollec111518/nq",
    },
)



class _ParameterStore:
    def __init__(self, key, value, depth, *args, **kwargs):
        self.__args = args
        self.__kwargs = kwargs
        self.key = key
        self.value = value
        self.depth = depth
        for key, value in self.__kwargs.items():
            setattr(self, key, value)

    def __getitem__(self, key):
        return self.__args[key]


def _forward_args_and_call(f1, f2, p):
    return (f.check_filter(p.key, p.value, p.depth) for f in [f1, f2])


class Filter:
    def __init__(self, func, name):
        self.func = func
        self.name = name
        self.args = []
        self.kwargs = {}

    def __call__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

    def __and__(self, f):
        return Filter(
            lambda p: all(_forward_args_and_call(self, f, p)), f"{self.name} & {f.name}"
        )

    def __or__(self, f):
        return Filter(
            lambda p: any(_forward_args_and_call(self, f, p)), f"{self.name} | {f.name}"
        )

    def check_filter(self, key, value, depth):
        all_args = _ParameterStore(key, value, depth, *self.args, **self.kwargs)
        try:
            return self.func(all_args)
        except Exception as e:
            return False

    def __repr__(self):
        return f"{self.name}({self.args}, {self.kwargs})"


def make_filter(func):
    def make_filter_wrapper(*fargs, **fkwargs):
        new_filter = Filter(func, func.__name__)
        new_filter(*fargs, **fkwargs)
        return new_filter

    return make_filter_wrapper


@make_filter
def get_all(box):
    return True


@make_filter
def get_index(box):
    return isinstance(box.key, int) and box.key == box[0]


@make_filter
def get_key(box):
    return box.key == box[0]


@make_filter
def get_range(box):
    return isinstance(box.key, int) and box[0] <= box.key < box[1]


@make_filter
def get_multiple(box):
    return box.key in box.args


@make_filter
def contains_attribute(box):
    return box[0] in box.value


@make_filter
def when_attribute(box):
    if not box[0] in box.value:
        return False
    if hasattr(box, "equals"):
        return box.value[box[0]] == box.equals
    elif hasattr(box, "greater_than"):
        return box.value[box[0]] > box.greater_than
    elif hasattr(box, "lesser_than"):
        return box.value[box[0]] < box.lesser_than
    elif hasattr(box, "check"):
        return box.check(box.value[box[0]])
